import Interface
from types import ClassType


class FeatureWeaving(Interface.Base):

    """Defines the weaving process for an attribute of a component class"""

    def weaveFeature(self,name,item,maker,verticalContext):
        
        """Notification that a feature needs weaving.

        * name -- the name of the feature to be woven
        
        * item -- the object bound to 'name' in the current source aspect layer

        * maker -- the ComponentMaker doing the weaving
        
        * verticalContext -- a function which returns the name of its argument
        in the source aspect layer, or the argument itself if it is not found.
        
        This message is sent when a FeatureDef is first detected (or defaulted)
        during weaving.  If your FeatureDef wants to be called in every
        subsequent layer when an attribute named 'name' is present, and at
        "finishing" time, it should install itself in the feature registry
        by executing::
        
         maker.features[name]=self

        If you do not do this, the weaveFeature() message will be the only one
        your FeatureDef receives.
        """









    def finishFeature(self,name,maker):
        
        """Last chance to affect the output of this feature.

        * name -- the name of the feature

        * maker -- the ComponentMaker doing the weaving

        This message is sent during the "finishing" phase of component
        construction, to all FeatureDefs which have registered themselves with
        the ComponentMaker.
        """


    def localDependencies(self,name,maker):
        
        """Return the names of other features this feature depends on.

        * name -- the name of the feature
        
        * maker -- the ComponentMaker doing the weaving

        Return: a sequence of names of features that must be finished before
        this feature is finished.

        This method is called by the ComponentMaker to build a dependency graph
        during the "finishing" phase.  Features are "finished" in order based on
        their local dependencies (i.e., order between attributes of the same
        class).
        """











    def globalDependencies(self,name,maker):
        """Return dependencies on features in the next outer class.

        * name -- the name of the feature
        
        * maker -- the ComponentMaker doing the weaving

        This method is used by certain features (notably the standard
        ClassFeature) to calculate their localDependencies.  It is not called
        by ComponentMakers directly.  This method is required to exist so that
        features which create an instance of a class that's being woven, can
        force their containing class to be "finished" after the class that the
        feature wishes to instantiate.  For example::

            class DiceService(Aspect):

                class RandomValueFeature:

                    def __init__(self,lo,hi):
                        self.lo = lo; self.hi = hi
                        
                    def next(self):
                        return randint(self.lo,self.hi)

                class DiceElement:
                    
                    diceValue = FeatureExpression('RandomValueFeature(1,6)')
                    
                    def roll(self):
                        return self.diceValue.next()


        When the ComponentMaker asks the ClassFeature responsible for the
        DiceElement class what its localDependencies are, it will check the
        globalDependencies() of its contained features.  Meanwhile, the
        FeatureExpression class has a globalDependencies() method that will
        return the tuple '("RandomValueFeature",)'.  Thus, the dependency graph
        for the classes will reflect DiceElement's dependency on
        RandomValueFeature, and they will be created in the right order at
        weaving time."""

isFeatureDef = FeatureWeaving.isImplementedBy

class FeatureDef:

    """Defines the weaving process for an attribute of a component class"""

    __implements__ = FeatureWeaving
    
    dependencies = ()
    
    def weaveFeature(self,name,item,maker,verticalContext):
        pass
        
    def finishFeature(self,name,maker):
        pass
    
    def localDependencies(self,name,maker):
        return self.dependencies

    def globalDependencies(self,name,maker):
        return self.localDependencies(name,maker)
        

class _OverwriteFeature(FeatureDef):

    """Feature definition for simple overwrite processing"""
    
    def weaveFeature(self,name,item,maker,verticalContext):
        maker[name]=item


OverwriteFeature = _OverwriteFeature()









class _ClassFeature(FeatureDef):

    def weaveFeature(self,name,item,maker,verticalContext):

        maker.features[name] = self
        ss = maker.get(name)
        assert (ss is None) or maker.isMaker(ss)

        if isAspect(item):
            if ss is None:
                ss = maker[name] = maker.childMaker(name)
            return item.weaveInto(ss,verticalContext)

        elif type(item) not in (ClassType,ExtensionClass):
            raise TypeError,("Incompatible feature types for member %s in component %s" % (name,maker.name))

        if ss is None:
            ss = maker[name] = maker.childMaker(name)
            
        ss.insertSupers(map(verticalContext, item.__bases__))
        ss.overlayDictionary(item.__dict__)

    def localDependencies(self,name,maker):
    
        ss = maker[name]
        d  = ss.dependencies.items()
        
        features = ss.features
        
        for name,feature in features.items():
            d.extend(list(feature.globalDependencies(name,ss)))
        return d        

    def globalDependencies(self,name,maker):
        return ()

    def finishFeature(self,name,maker):
        maker[name] = maker[name].finishComponent()

ClassFeature = _ClassFeature()

class FeatureExpression(FeatureDef):

    def __init__(self,expr=''):
        self.expr = expr
        self._compile(expr)
        
    def weaveFeature(self,name,item,maker,verticalContext):
        if isinstance(item,FeatureDef) and item is not self:
            item.weaveFeature(name,item,maker,verticalContext)
        else:
            maker.features[name] = self

    def finishFeature(self,name,maker):
        feature = eval(self.expr, maker.globals, maker.output)
        maker[name] = feature

    def __repr__(self): return self.expr


    def _compile(self,expr):
        code = compile(expr,"<string>","eval") # XXX find caller for file/line?
        names = code.co_names
        code = code.co_code
        deps = {}
        l = len(code); i=0
        while (i<l):
            c=ord(code[i])
            if c==101:  # LOAD_NAME
                deps[names[ord(code[i+1])+256*ord(code[i+2])]] = 1
                i=i+3
            elif c>=90:
                i=i+3   # HAVE_ARGUMENT
            else:
                i=i+1
        self.dependencies = tuple(deps.keys())






class Transform:

    def begin(self,maker,dict,verticalContext):
        pass
        
    def transform(self,maker,dict,verticalContext):
        pass
        
    def finishTransform(self,maker):
        pass

    
class FeatureSet(Transform):

    def __init__(self,*transforms,**namedFeatures):
        self.transforms = transforms
        self.namedFeatures = namedFeatures

    def begin(self,maker,dict,verticalContext):
        maker.features.update(self.namedFeatures)     # XXX check before overwrites?
        for t in self.transforms: t.begin(maker,dict,verticalContext)

    def transform(self,maker,dict,verticalContext):
        for t in self.transforms: t.transform(maker,dict,verticalContext)
        
    def finishTransform(self,maker):
        for t in self.transforms: t.finishTransform(maker)














def _DefaultFeatureMap(self, item):

    if isFeatureDef(item):
        return item
        
    if type(item) in (ClassType,ExtensionClass) or isAspect(item):
        return ClassFeature

    return OverwriteFeature
            


from TW.Aspects import isAspect, ExtensionClass

